#----------------------------------------------------------------------------------------------------------------------
#   
#	alindrv.py file
#
#	File used to read SDB information form spec card
#	History:
#   30/03/2015 - file created
#
#----------------------------------------------------------------------------------------------------------------------
__author__ = "Manolo Broseta"
__copyright__ = "Copyright 2015, ALBA"
__license__ = "GPLv3 or later"
__version__ = "1.1"
__email__ = "mbroseta@cells.es"
__status__ = "Development"

from spec_libc import *

import logging
import logging.handlers

import subprocess


import os


__OFFSET__ = 0x00 #0x100
__SDB_MAGIC__ = 0x5344422d

__SDB_RECORD_INTERCONNECT__ = 0x00
__SDB_RECORD_DEVICE__ = 0x01
__SDB_RECORD_BRIDGE__ = 0x02
__SDB_RECORD_INTEGRATION__ = 0x80
__SDB_RECORD_REPO_URL__ = 0x81
__SDB_RECORD_SYNTHESIS__ = 0x82
__SDB_RECORD_EMPTY__ = 0xff

__DEBUG__ = False	

class alinSDB():
	def __init__(self, offset=__OFFSET__):
		self.spec = Spec()
		self.init_offset = offset
		self.sdb_structure = []
		
		self._debugLevel = __DEBUG__
		
		try:
			subprocess.call(['modprobe -a','spec.ko'])
		except:
			pass
		
		# Logging
		self.__logging_folder = os.path.abspath(os.path.dirname(__file__))+"/logs/"
		self.__logging_file = self.__logging_folder+"alin.log"
		
		if not os.path.exists(self.__logging_folder):
			os.makedirs(self.__logging_folder)
		
		self._devlogger = logging.getLogger('alinSDB')
		if not len(self._devlogger.handlers):
			self._devlogger.setLevel(logging.DEBUG)
		
			handler = logging.handlers.RotatingFileHandler(self.__logging_file, maxBytes=5000000, backupCount=5)
			handler.setLevel(logging.INFO)
		
			formatter = logging.Formatter('%(asctime)s - %(name)s - %(message)s')
			handler.setFormatter(formatter)
			self._devlogger.addHandler(handler)
		
	def SDB_setLogMessage(self,msg):
		self._devlogger.info(msg)
		if self._debugLevel: print msg		

	def SDB_setLogEnable(self,dbg=False):
		self._debugLevel = dbg
		self.setLogMessage("SDB_setLogEnable::Debug level set to %s"%self._debugLevel)		

	def SDB_readData(self, address=0):
		self.SDB_setLogMessage("SDB_readData:: Address %s"%str(address))
		self.base_address = address
		if self.SDB_check():		
			self.device_bridges = [0]
			self.sdb_structure = []
			self.devices_counter = 0
			bus_level = 0
			
			# Get devices in every block of devicesper bridge
			while len(self.device_bridges)>0:
				bridge_pos = self.device_bridges.pop()
				
				if bridge_pos!=0:
					add_data = self.sdb_structure[bridge_pos]['sdb_record_type']
					self.base_address += self.SDB_getAddress(add_data)
					bus_level +=1
					
				devs = self.SDB_getDevices(self.base_address)
				if devs>0:
					for el in range(devs):
						subaddress = el*64
						dev_address = self.base_address + subaddress
						self.SDB_readBlock(el, bus_level, dev_address)
						self.devices_counter +=1
		self.SDB_setLogMessage("SDB_readData:: End")
		
	def SDB_getDevices(self, address):
		# Get number of devices #
		rdata = self.spec.specReadL(address+4, hexformat=False)
		devices = int((rdata&0xffff0000)>>16)
		self.SDB_setLogMessage("SDB_getDevices:: %s decives found in address %s"%(str(devices),hex(address)))
		return devices
		
	def SDB_getAddress(self,*args):
		data = args[0]
		address = 0
		for el in data:
			address = (address<<8) + el
		return address
	
	def SDB_readAddress(self,address):
		if address % 4:
			address -= (address % 4)
		rdata = self.spec.specReadL(address, hexformat=False)
		self.SDB_setLogMessage("SDB_readAddress:: Address=%s Value=%s"%(hex(address),hex(rdata)))
		return rdata
		
	def SDB_writeAddress(self,address, data_w=0):
		if address % 4:
			address -= (address % 4)
		self.spec.specWriteL(address, data_w)
		self.SDB_setLogMessage("SDB_writeAddress:: Address=%s Value=%s"%(hex(address),hex(data_w)))
		self.SDB_readAddress(address)
		
	def SDB_readBlock(self, dev=0, bus=0, address=0):
		data_block = []
		for i in range (0,16):
			data = self.spec.specReadL(address+(i*4),hexformat=False)
			temp = (data&0xff000000)>>24
			data_block.append(temp)
			temp = (data&0x00ff0000)>>16
			data_block.append(temp)
			temp = (data&0x0000ff00)>>8
			data_block.append(temp)
			temp = (data&0x000000ff)
			data_block.append(temp)

		sdb_component = {}
		sdb_component['dev_num'] = self.devices_counter
		sdb_component['bus'] = bus
		sdb_component['devnum_in_bus'] = dev
		sdb_component['base_address'] = self.base_address
		sdb_component['dev_address'] = address
		sdb_component['interconnect'] = data_block[63]
		
		if sdb_component['interconnect'] == __SDB_RECORD_REPO_URL__:
			# REPOSITORY-URL RECORD TYPE
			sdb_component['repo_url'] = [(data_block[i]) for i in range(0,63)]
		elif sdb_component['interconnect'] == __SDB_RECORD_SYNTHESIS__:
			# SYNTHESIS RECORD TYPE
			sdb_component['name'] = [(data_block[i]) for i in range(0,16)]
			sdb_component['commit_id'] = [(data_block[i]) for i in range(16,32)]
			sdb_component['tool_name'] = [(data_block[i]) for i in range(32,40)]
			sdb_component['tool_ver'] = [(data_block[i]) for i in range(40,44)]
			sdb_component['date'] = [(data_block[i]) for i in range(44,48)]
			sdb_component['usr_name'] = [(data_block[i]) for i in range (48,63)]			
		else:
			# REST OF RECORD DEVICES HAVE SIMILAR SDB STRUCTURE
			sdb_component['sdb_record_type'] = [(data_block[i]) for i in range(0,8)]
			if sdb_component['interconnect'] != __SDB_RECORD_INTEGRATION__:
				sdb_component['first_address'] = [(data_block[i]) for i in range(8,16)]
				sdb_component['last_address'] = [(data_block[i]) for i in range(16,24)]
			sdb_component['vendor'] = [(data_block[i]) for i in range(24,32)]
			sdb_component['device'] = [(data_block[i]) for i in range(32,36)]
			sdb_component['version'] = [(data_block[i]) for i in range(36,40)]
			sdb_component['date'] = [(data_block[i]) for i in range(40,44)]
			sdb_component['name'] = [(data_block[i]) for i in range (44,63)]

			if sdb_component['interconnect'] == __SDB_RECORD_BRIDGE__:
				self.device_bridges.append(self.devices_counter)
		
		self.sdb_structure.append(sdb_component)
		self.SDB_setLogMessage("SDB_readBlock:: %s record type found at bus level %s address %s"%(hex(sdb_component['interconnect']), str(bus), hex(address)))
		
	def SDB_getData(self):
		self.SDB_setLogMessage("SDB_getData:: SDB structure read")
		return self.sdb_structure
					
	def SDB_showTree(self):
		self.SDB_setLogMessage("SDB_showTree:: SDB tree structure show")
		print "#\tBusPath\tVendorId\tProduct\t\tAddress range(hex)\tDescription"

		for el in self.sdb_structure:
			add_val = "..........."
			if el['interconnect']==__SDB_RECORD_REPO_URL__:
				vendor_device = "..<< Repository-url >>.."
				name = ''.join([chr(a) for a in el['repo_url']])
			elif el['interconnect']== __SDB_RECORD_SYNTHESIS__:				
				vendor_device = "....<< Synthesis >>....."
				name = ''.join([chr(a) for a in el['name']])
			elif el['interconnect']== __SDB_RECORD_INTEGRATION__:
				vendor = ''.join(['%02x'%a for a in el['vendor']])
				device = ''.join(['%02x'%a for a in el['device']])
				vendor_device = vendor+":"+device
				name = ''.join([chr(a) for a in el['name']])
			else:
				first_address = self.SDB_getAddress(el['first_address'])
				last_address = self.SDB_getAddress(el['last_address'])
				add_val = hex(first_address+el['base_address'])+"-"+hex(last_address+el['base_address'])
				vendor = ''.join(['%02x'%a for a in el['vendor']])
				device = ''.join(['%02x'%a for a in el['device']])
				vendor_device = vendor+":"+device
				name = ''.join([chr(a) for a in el['name']])
			print "%s\t%s.%s\t%s\t%s\t\t%s"%(str(el['dev_num']),str(el['bus']),str(el['devnum_in_bus']),vendor_device,add_val,name)

	def SDB_showInfo(self):
		self.SDB_setLogMessage("SDB_showInfo:: SDB info structure show")
		for el in self.sdb_structure:
			print "BaseAdress: \t\t%s"%hex(el['base_address'])
			print "BusPath level:  \t%s.%s"%(el['bus'],el['devnum_in_bus'])

			if el['interconnect'] == __SDB_RECORD_INTEGRATION__: #Informative integration structure#
				print "Record type\t\tIntegration Record"
			elif el['interconnect'] == __SDB_RECORD_REPO_URL__: #Empty record
				print "Record type\t\tRepository-Url"
				print "Url\t\t\t0x00: %s"%''.join([chr(a) for a in el['repo_url']])
				print "Record Type\t\t0x3f: %s\n"%hex(el['interconnect'])
				continue
			elif el['interconnect'] == __SDB_RECORD_SYNTHESIS__: #Empty record
				print "Record type\t\tSynthesis Record"
				print "Project \t\t0x00: %s"%''.join([chr(a) for a in el['name']])
				print "Commit_id\t\t0x10: %s"%' '.join(['%02x'%a for a in el['commit_id']])
				print "Tool name\t\t0x00: %s"%''.join([chr(a) for a in el['tool_name']])
				print "Tool ver\t\t0x00: %s"%' '.join(['%02x'%a for a in el['tool_ver']])
				print "Date\t\t\t0x28: %s"%' '.join(['%02x'%a for a in el['date']])
				print "User name\t\t0x00: %s"%''.join([chr(a) for a in el['usr_name']])
				print "Record Type\t\t0x3f: %s\n"%hex(el['interconnect'])
				continue
			elif el['interconnect'] == __SDB_RECORD_INTERCONNECT__ : #SDB Interconnect#/
				print "Record type\t\tSDB Interconnect"
				print "Magic 'SDB-'\t\t0x00: %s"%' '.join(['%02x'%(el['sdb_record_type'][i]) for i in range(0,4)])
				print "Number of Records\t0x04: %s"%' '.join(['%02x'%(el['sdb_record_type'][i]) for i in range(4,6)])
				print "SDB Version\t\t0x06: %02x"%el['sdb_record_type'][6]
				print "Bus type\t\t0x07: %02x"%el['sdb_record_type'][7]
			elif el['interconnect'] == __SDB_RECORD_DEVICE__: #Device definition#/
				print "Record type\t\tDevice Record"
				print "ABI class\t\t0x00: %s"%' '.join(['%02x'%(el['sdb_record_type'][i]) for i in range(0,2)])
				print "ABI version major\t0x04: %02x"%el['sdb_record_type'][2]
				print "ABI version minor\t0x06: %02x"%el['sdb_record_type'][3]
				print "Bus-specific field\t0x07: %s"%' '.join(['%02x'%(el['sdb_record_type'][i]) for i in range(4,8)])
				print "First address\t\t0x08: %s"%' '.join(['%02x'%a for a in el['first_address']])
				print "Last address\t\t0x10: %s"%' '.join(['%02x'%a for a in el['last_address']])

			elif el['interconnect'] == __SDB_RECORD_BRIDGE__: #Bridge to sub-bus definition#/
 				print "Record type\t\tBridge Record"
				print "SDB Child\t\t0x00: %s"%' '.join(['%02x'%(el['sdb_record_type'][i]) for i in range(0,8)])
				print "First address\t\t0x08: %s"%' '.join(['%02x'%a for a in el['first_address']])
				print "Last address\t\t0x10: %s"%' '.join(['%02x'%a for a in el['last_address']])
			elif el['interconnect'] == __SDB_RECORD_EMPTY__: #Empty record
				# reserved #
				continue

			print "Vendor  \t\t0x18: %s"%' '.join(['%02x'%a for a in el['vendor']])
			print "Device  \t\t0x20: %s"%' '.join(['%02x'%a for a in el['device']])
			print "Version \t\t0x24: %s"%' '.join(['%02x'%a for a in el['version']])
			print "Date\t\t\t0x28: %s"%' '.join(['%02x'%a for a in el['date']])
			print "Name\t\t\t0x2c: %s"%''.join([chr(a) for a in el['name']])
			print "Record Type\t\t0x3f: %s\n"%hex(el['interconnect'])
		
	def SDB_check(self):
		# Read SDB Magic #
		rdata = self.spec.specReadL(self.base_address, hexformat=False)
		if (rdata==__SDB_MAGIC__):
			self.SDB_setLogMessage("SDB_check:: SDB Magic number found at %s address"%hex(self.base_address))
			return True
		else:
			self.SDB_setLogMessage("SDB_check:: SDB Magic not found!!")
		return False
	
	def SDB_loadFile(self,filename):
		self.filename = filename
		#val = self.spec.specLoadBitstream(self.filename)	
		temp = 'cp -f '+self.filename+' /lib/firmware/fmc'
		os.system(temp)
		os.system('rmmod spec')
		temp = 'modprobe spec fw_name="/fmc/'+self.filename.split("/")[-1]+'" show_sdb=2'
		os.system(temp)
		os.system('dmesg | grep spec')
		val = True
		if val:
			self.SDB_setLogMessage("SDB_loadFile:: File %s loaded succesfully"%filename)
			# Reload data after loading the new file
			self.SDB_readData(self.init_offset)
		else:
			self.SDB_setLogMessage("SDB_loadFile:: Problems loading file %s"%filename)
		
		return val
		
	def SDB_readAddressRange(self,init_address, end_address):
		if init_address % 4:
			init_address -= (init_address % 4)
		if end_address % 16:
			end_address += (15 - (end_address % 16))
		block = []
		line = []
		for add in range(init_address, end_address+4, 4):
			if add % 16  == 0:
				line.append(add)
			rdata = self.spec.specReadL(add, hexformat=False)
			temp = (rdata&0xff000000)>>24
			line.append(temp)
			temp = (rdata&0x00ff0000)>>16
			line.append(temp)
			temp = (rdata&0x0000ff00)>>8
			line.append(temp)
			temp = (rdata&0x000000ff)
			line.append(temp)
			if add % 16 == 12:
				block.append(line)
				line = []
		self.SDB_setLogMessage("SDB_readAddressRange:: Read address range from %s to %s"%(hex(init_address), hex(end_address+4)))				
		return block

	#def SDB_showDeviceList(self):
		#self.ct_el = []
		#ct = 0
		#self.SDB_setLogMessage("SDB_showDeviceList:: Show Device List")
		#for el in self.sdb_structure:
			#if (el['interconnect'] == __SDB_RECORD_BRIDGE__) or\
				#(el['interconnect'] == __SDB_RECORD_INTERCONNECT__) or\
				#(el['interconnect'] == __SDB_RECORD_DEVICE__):
				#print "\t(%s): %s"%(str(ct),''.join([chr(a) for a in el['name']]))
				#ct += 1
				#self.ct_el.append(el['dev_num'])
			
	#def SDB_showDeviceMemory(self,devnum):
		#self.SDB_setLogMessage("SDB_showDeviceMemory:: Show Device Memoryr for device number %s"%str(devnum))
		#if  len(self.ct_el)>int(devnum):
			#el = self.sdb_structure[int(self.ct_el[int(devnum)])]
			#init_add = self.SDB_getAddress(el['first_address'])+el['base_address']
			#end_add =  self.SDB_getAddress(el['last_address'])+el['base_address']
			#print "\nVendor: %s"%''.join(['%02x'%a for a in el['vendor']])
			#print "Device: %s"%''.join(['%02x'%a for a in el['device']])
			#print "Name: %s"%''.join([chr(a) for a in el['name']])
			#print "First address=%s"%hex(init_add)
			#print "Last address=%s\n"%hex(end_add)
			#block = []
			#block = self.SDB_readAddressRange(init_add,end_add)
			#return block
		#else:
			#self.SDB_setLogMessage("SDB_showDeviceMemory:: Nothing to show")
			#return None
			
	def SDB_showDeviceMemory(self,devnum):
		self.SDB_setLogMessage("SDB_showDeviceMemory:: Show Device Memoryr for device number %s"%str(devnum))
		if  len(self.sdb_structure)>int(devnum):
			el = self.sdb_structure[int(devnum)]
			if (el['interconnect'] == __SDB_RECORD_BRIDGE__) or\
				(el['interconnect'] == __SDB_RECORD_INTERCONNECT__) or\
				(el['interconnect'] == __SDB_RECORD_DEVICE__):
					init_add = self.SDB_getAddress(el['first_address'])+el['base_address']
					end_add =  self.SDB_getAddress(el['last_address'])+el['base_address']
					print "\nVendor: %s"%''.join(['%02x'%a for a in el['vendor']])
					print "Device: %s"%''.join(['%02x'%a for a in el['device']])
					print "Name: %s"%''.join([chr(a) for a in el['name']])
					print "First address=%s"%hex(init_add)
					print "Last address=%s\n"%hex(end_add)
					block = []
					block = self.SDB_readAddressRange(init_add,end_add)
					return block
			else:
				print "** Not possible to show device map for this device **"
				return None
		else:
			self.SDB_setLogMessage("SDB_showDeviceMemory:: Nothing to show")
			return None
		
	def SDB_getDeviceMemory(self,devname="",vend="", prod=""):
		self.SDB_setLogMessage("SDB_getDeviceMemory:: Read device  Memory for device vendor=%s prodcut=%s"%(hex(vend),hex(prod)))
		for el in self.sdb_structure:
			if (el['interconnect'] == __SDB_RECORD_BRIDGE__) or\
				(el['interconnect'] == __SDB_RECORD_INTERCONNECT__) or\
				(el['interconnect'] == __SDB_RECORD_DEVICE__):
				name = ''.join([chr(a) for a in el['name']])
				vendor = ''.join(['%02x'%a for a in el['vendor']])
				device = ''.join(['%02x'%a for a in el['device']])		
				if (devname != "" and devname.lower() in name.lower()) or \
					((vend != "" and vend == int(vendor,16)) and \
					(prod != "" and prod == int(device,16))):
					data = {}
					init_add = self.SDB_getAddress(el['first_address'])+el['base_address']
					end_add =  self.SDB_getAddress(el['last_address'])+el['base_address']
					
					if init_add % 4:
						init_add -= (init_add % 4)
					if end_add % 4:
						end_add -= (end_add % 4)
					block = []
					for add in range(init_add, end_add+4, 4):
						rdata = self.spec.specReadL(add, hexformat=False)
						block.append(rdata)			
					data['init_add'] = init_add
					data['end_add'] = end_add
					data['data'] = block
					return data
		self.SDB_setLogMessage("SDB_getDeviceMemory:: Nothing to show")
		return None


