#----------------------------------------------------------------------------------------------------------------------
#   
#	alindev.py file
#
#	File containing the device class, to access device registers
#	History:
#   29/07/2015 - file created
#
#----------------------------------------------------------------------------------------------------------------------
__author__ = "Manolo Broseta"
__copyright__ = "Copyright 2015, ALBA"
__license__ = "GPLv3 or later"
__version__ = "1.0"
__email__ = "mbroseta@cells.es"
__status__ = "Development"

import os.path, pickle
import logging, logging.handlers

from alin import *

__OFFSET__ = 0x00 #0x100
__SDB_RECORD_INTERCONNECT__ = 0x00
__SDB_RECORD_DEVICE__ = 0x01
__SDB_RECORD_BRIDGE__ = 0x02
__SDB_RECORD_INTEGRATION__ = 0x80
__SDB_RECORD_REPO_URL__ = 0x81
__SDB_RECORD_SYNTHESIS__ = 0x82
__SDB_RECORD_EMPTY__ = 0xff

__DEBUG__ = False

class alinDevice():
	def __init__(self,device=None):
		self._debugLevel = __DEBUG__
		self._devMap = {}
		self.sdbDrv = alinSDB(__OFFSET__)
		self.sdbDrv.SDB_readData(__OFFSET__)
		
		self.__device_folder = os.path.abspath(os.path.dirname(__file__))+"/deviceslib/"
		
		# Logging
		self.__logging_file = os.path.abspath(os.path.dirname(__file__))+"/logs/alin.log"
		self._devlogger = logging.getLogger('alinDevice')
		if not len(self._devlogger.handlers):
			self._devlogger.setLevel(logging.DEBUG)
			
			handler = logging.handlers.RotatingFileHandler(self.__logging_file, maxBytes=5000000, backupCount=5)
			handler.setLevel(logging.INFO)
			
			formatter = logging.Formatter('%(asctime)s - %(name)s - %(message)s')
			handler.setFormatter(formatter)
			self._devlogger.addHandler(handler)
		
		if device:
			self.setDevice(device)
		
	def setDevice(self,device):
		if ":" not in device:
			self.setLogMessage("Incorret device name format!"%self._device.upper())
			return
		self.vendorId, self.product = device.split(":")
		self.vendorId = int(self.vendorId.lower(),16)
		self.product = int(self.product.lower(),16)
		self._device = "%016x:%08x"%(self.vendorId,self.product)

		temp = []
		temp = self.sdbDrv.SDB_getData()
		for el in temp:
			if (el['interconnect'] == __SDB_RECORD_BRIDGE__) or\
				(el['interconnect'] == __SDB_RECORD_INTERCONNECT__) or\
				(el['interconnect'] == __SDB_RECORD_DEVICE__):
				vendor = ''.join(['%02x'%a for a in el['vendor']])
				device = ''.join(['%02x'%a for a in el['device']])		
				if  ((self.vendorId == int(vendor,16)) and  self.product == int(device,16)):
					self.initAddress = self.sdbDrv.SDB_getAddress(el['first_address'])+el['base_address']
					
		pickfile = self.__device_folder+self._device
		self.setLogMessage("setDevice::Reading %s pickle file"%pickfile)
		if not os.path.isfile(pickfile):
			self.setLogMessage("setDevice::%s device register map not found!!"%self._device.upper())
			self._device = None
		else:
			self.setLogMessage("setDevice::%s device found!"%self._device.upper())
		
		self._devMap = {}
		try:
			with open(pickfile) as output:
				self._devMap = pickle.load(output)
		except Exception, e:
			self.setLogMessage("setDevice::Problem found while reading %s"%pickfile)		
				
	def setLogMessage(self,msg):
		self._devlogger.info(msg)		
		if self._debugLevel: print msg
		
	def setLogEnable(self,dbg=False):
		self._debugLevel = dbg
		self.setLogMessage("setDebugLevel::Debug level set to %s"%self._debugLevel)		

	def getDeviceInfo(self):
		self.setLogMessage("getDeviceInfo::")		
		return self._devMap
	
	def getDeviceName(self):
		if self._devMap and "name" in self._devMap.keys():
			self.setLogMessage("getDeviceName::")		
			return self._devMap["name"]
		else:
			self.setLogMessage("getDeviceName::Name not availble")		
		return None		
	
	def getDeviceDescription(self):
		if self._devMap and "description" in self._devMap.keys():
			self.setLogMessage("getDeviceDescription::")		
			return self._devMap["description"]
		else:
			self.setLogMessage("getDeviceDescription::Description not availble")		
		return None

	def getDeviceData(self):
		valueList = []
		rd_data = {}
		rd_data = self.sdbDrv.SDB_getDeviceMemory(vend=self.vendorId, prod=self.product)
		if rd_data and self._devMap:
			for el in self._devMap['regs']:
				regname = el.keys()[0]
				address = el[el.keys()[0]]['address']
				position = el[el.keys()[0]]['bit_position']
				size = el[el.keys()[0]]['size']
				
				mask = (2**size - 1) << position
				
				value = rd_data['data'][address]
				value = (value&mask)>>position
				
				valueList.append((regname.upper(),hex(value)))
			self.setLogMessage("getDeviceData::Read all device data list")
		else:
			self.setLogMessage("getDeviceData:Not possible to get the data from Memory!!")
		
		return valueList

	def getAttributesList(self):
		valueList = []
		if self._devMap:
			valueList = [el.keys()[0] for el in self._devMap['regs']]
			self.setLogMessage("getAttributesList::Attribute list read")
		else:
			self.setLogMessage("getAttributesList::No data found")
		return valueList

	def writeAttribute(self,attrName,attrVal):
		if attrName is not None and self._devMap:
			for el in self._devMap['regs']:
				if attrName.lower() == el.keys()[0].lower():
					address = el[el.keys()[0]]['address']
					sdb_address = self.initAddress+(address*4)
					position = el[el.keys()[0]]['bit_position']
					size = el[el.keys()[0]]['size']
					access = el[el.keys()[0]]['access'].split()[0]

					mask = (2**size - 1) << position
					
					value = self.sdbDrv.SDB_readAddress(sdb_address)

					if access == "WRITE_ONLY" or access == "READ_WRITE":
						inv_mask = 0xffffffff - mask
						value = value & inv_mask
						data = (attrVal&(2**size - 1)) << position
						value = (value | data)
						
						self.sdbDrv.SDB_writeAddress(sdb_address,value)
						
						self.setLogMessage("writeAttribute::Attribute %s address: %s value=%s"%(attrName, hex(sdb_address), hex(value)))
					else:
						self.setLogMessage("writeAttribute::Attribute is %s type"%access)
		else:
			self.setLogMessage("writeAttribute::Cannot write %s attribute"%attrName)

	def readAttribute(self,attrName):
		value = None
		if attrName is not None and self._devMap:
			for el in self._devMap['regs']:
				if attrName.lower() == el.keys()[0].lower():
					address = el[el.keys()[0]]['address']
					sdb_address = self.initAddress+(address*4)
					position = el[el.keys()[0]]['bit_position']
					size = el[el.keys()[0]]['size']
					access = el[el.keys()[0]]['access'].split()[0]

					mask = (2**size - 1) << position
					value = self.sdbDrv.SDB_readAddress(sdb_address)
					value = (value&mask)>>position
					
					self.setLogMessage("readAttribute::Attribute %s address: %s value=%s"%(attrName, hex(sdb_address), hex(value)))
		else:
			self.setLogMessage("readAttribute::Cannot read %s attribute"%attrName)
		return value
	
	def getAttributeInfo(self,attrName):
		valueDict = {}
		if attrName is not None and self._devMap:
			for el in self._devMap['regs']:
				if attrName.lower() == el.keys()[0].lower():
					address = el[el.keys()[0]]['address']
					sdb_address = self.initAddress+(address*4)
					position = el[el.keys()[0]]['bit_position']
					size = el[el.keys()[0]]['size']
					access = el[el.keys()[0]]['access'].split()[0]
					desc = el[el.keys()[0]]['description']

					mask = (2**size - 1) << position
					value = self.sdbDrv.SDB_readAddress(sdb_address)
					value = (value&mask)>>position
					
					valueDict['name'] = attrName
					valueDict['desc'] = desc
					valueDict['address'] = sdb_address
					valueDict['position'] = position
					valueDict['size'] = size
					valueDict['access'] = access
					valueDict['value'] = value
			self.setLogMessage("getAttributeInfo::Attribute infop dict read")
		else:
			self.setLogMessage("getAttributeInfo::Cannot get %s attribute info"%attrName)
		return valueDict

